@@ -2,10 +2,12 @@ import os
 import sqlite3
 import stat
 import time
-from typing import Any, Callable, Dict, Optional, Union
+import uuid
+from abc import ABC, ABCMeta, abstractmethod
+from typing import Any, Callable, Dict, Optional
 
 from jupyter_core.paths import jupyter_data_dir
-from traitlets import Int, TraitError, Unicode, validate
+from traitlets import TraitError, Unicode, validate
 from traitlets.config.configurable import LoggingConfigurable
 
 
@@ -37,15 +39,20 @@ def log(log_before, log_after):
     return decorator
 
 
-class BaseFileIdManager(LoggingConfigurable):
+class FileIdManagerMeta(ABCMeta, type(LoggingConfigurable)):  # type: ignore
+    pass
+
+
+class BaseFileIdManager(ABC, LoggingConfigurable, metaclass=FileIdManagerMeta):
     """
     Base class for File ID manager implementations. All File ID
     managers should inherit from this class.
     """
 
     root_dir = Unicode(
-        help=("The root directory being served by Jupyter server. Must be an absolute path."),
+        help="The root directory being served by Jupyter server.",
         config=False,
+        allow_none=True,
     )
 
     db_path = Unicode(
@@ -57,44 +64,164 @@ class BaseFileIdManager(LoggingConfigurable):
         config=True,
     )
 
-    @validate("root_dir", "db_path")
-    def _validate_abspath_traits(self, proposal):
+    @validate("db_path")
+    def _validate_db_path(self, proposal):
         if proposal["value"] is None:
-            raise TraitError(f"FileIdManager : {proposal['trait'].name} must not be None")
+            raise TraitError(f"BaseFileIdManager : {proposal['trait'].name} must not be None")
         if not os.path.isabs(proposal["value"]):
-            raise TraitError(f"FileIdManager : {proposal['trait'].name} must be an absolute path")
+            raise TraitError(
+                f"BaseFileIdManager : {proposal['trait'].name} must be an absolute path"
+            )
         return proposal["value"]
 
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    @staticmethod
+    def _uuid() -> str:
+        return str(uuid.uuid4())
 
-    def index(self, path: str) -> Union[int, str, None]:
-        raise NotImplementedError("must be implemented by subclass")
+    def _normalize_path(self, path: str) -> str:
+        """Accepts an API path and returns a filesystem path, i.e. one prefixed
+        by root_dir and uses os.path.sep."""
+        # use commonprefix instead of commonpath, since root_dir may not be a
+        # absolute POSIX path.
+        if os.path.commonprefix([self.root_dir, path]) != self.root_dir:
+            path = os.path.join(self.root_dir, path)
 
-    def get_id(self, path: str) -> Union[int, str, None]:
-        raise NotImplementedError("must be implemented by subclass")
+        return path
 
-    def get_path(self, id: Union[int, str]) -> Union[int, str, None]:
-        raise NotImplementedError("must be implemented by subclass")
+    def _from_normalized_path(self, path: Optional[str]) -> Optional[str]:
+        """Accepts a filesystem path and returns an API path, i.e. one relative
+        to root_dir and uses forward slashes as the path separator. Returns
+        `None` if the given path is None or is not relative to root_dir."""
+        if path is None:
+            return None
 
-    def move(self, old_path: str, new_path: str) -> Union[int, str, None]:
-        raise NotImplementedError("must be implemented by subclass")
+        if os.path.commonprefix([self.root_dir, path]) != self.root_dir:
+            return None
 
-    def copy(self, from_path: str, to_path: str) -> Union[int, str, None]:
-        raise NotImplementedError("must be implemented by subclass")
+        relpath = os.path.relpath(path, self.root_dir)
+        # always use forward slashes to delimit children
+        relpath = relpath.replace(os.path.sep, "/")
 
+        return relpath
+
+    def _move_recursive(self, old_path: str, new_path: str, sep: str = os.path.sep) -> None:
+        """Move all children of a given directory at `old_path` to a new
+        directory at `new_path`, delimited by `sep`."""
+        old_path_glob = old_path + sep + "*"
+        records = self.con.execute(
+            "SELECT id, path FROM Files WHERE path GLOB ?", (old_path_glob,)
+        ).fetchall()
+
+        for record in records:
+            id, old_recpath = record
+            new_recpath = os.path.join(new_path, os.path.relpath(old_recpath, start=old_path))
+            self.con.execute("UPDATE Files SET path = ? WHERE id = ?", (new_recpath, id))
+
+    def _copy_recursive(self, from_path: str, to_path: str, sep: str = os.path.sep) -> None:
+        """Copy all children of a given directory at `from_path` to a new
+        directory at `to_path`, delimited by `sep`."""
+        from_path_glob = from_path + sep + "*"
+        records = self.con.execute(
+            "SELECT path FROM Files WHERE path GLOB ?", (from_path_glob,)
+        ).fetchall()
+
+        for record in records:
+            (from_recpath,) = record
+            to_recpath = os.path.join(to_path, os.path.relpath(from_recpath, start=from_path))
+            self.con.execute(
+                "INSERT INTO Files (id, path) VALUES (?, ?)", (self._uuid(), to_recpath)
+            )
+
+    def _delete_recursive(self, path: str, sep: str = os.path.sep) -> None:
+        """Delete all children of a given directory, delimited by `sep`."""
+        path_glob = path + sep + "*"
+        self.con.execute("DELETE FROM Files WHERE path GLOB ?", (path_glob,))
+
+    @abstractmethod
+    def index(self, path: str) -> Optional[str]:
+        """Returns the file ID for the file corresponding to `path`.
+
+        If `path` is not already indexed, a new file ID will be created and associated
+        with `path`, otherwise the existing file ID will be returned. Returns None if
+        `path` does not correspond to an object as determined by the implementation.
+        """
+        pass
+
+    @abstractmethod
+    def get_id(self, path: str) -> Optional[str]:
+        """Retrieves the file ID associated with the given file path.
+
+        Returns None if the file has not yet been indexed.
+        """
+        pass
+
+    @abstractmethod
+    def get_path(self, id: str) -> Optional[str]:
+        """
+        Accepts a file ID and returns the API path to that file. Returns None if
+        the file ID does not exist.
+
+        Notes
+        -----
+        - See `_from_normalized_path()` for implementation details on how to
+        convert a filesystem path to an API path.
+        """
+        pass
+
+    @abstractmethod
+    def move(self, old_path: str, new_path: str) -> Optional[str]:
+        """Emulates file move operations by updating the old file path to the new file path.
+
+        If old_path corresponds to a directory (as determined by the implementation), all indexed
+        file paths prefixed with old_path will have their locations updated and prefixed with new_path.
+
+        Returns the file ID if new_path is valid, otherwise None.
+        """
+        pass
+
+    @abstractmethod
+    def copy(self, from_path: str, to_path: str) -> Optional[str]:
+        """Emulates file copy operations by copying the entry corresponding to from_path
+         and inserting an entry corresponding to to_path.
+
+        If from_path corresponds to a directory (as determined by the implementation), all indexed
+        file paths prefixed with from_path will have their entries copying and inserted to entries
+        corresponding to to_path.
+
+        Returns the file ID if to_path is valid, otherwise None.
+        """
+        pass
+
+    @abstractmethod
     def delete(self, path: str) -> None:
-        raise NotImplementedError("must be implemented by subclass")
+        """Emulates file delete operations by deleting the entry corresponding to path.
 
-    def save(self, path: str) -> None:
-        raise NotImplementedError("must be implemented by subclass")
+        If path corresponds to a directory (as determined by the implementation), all indexed
+        file paths will have their entries deleted.
 
+        Returns None.
+        """
+        pass
+
+    @abstractmethod
+    def save(self, path: str) -> Optional[str]:
+        """Emulates file save operations by inserting the entry corresponding to path.
+
+        Entries are inserted when one corresponding to path does not already exist.
+
+        Returns the ID corresponding to path or None if path is determined to not be valid.
+        """
+        pass
+
+    @abstractmethod
     def get_handlers_by_action(self) -> Dict[str, Optional[Callable[[Dict[str, Any]], Any]]]:
-        """Returns a dictionary whose keys are contents manager event actions
-        and whose values are callables invoked upon receipt of an event of the
-        same action. The callable accepts the body of the event as its only
-        argument. To ignore an event action, set the value to `None`."""
-        raise NotImplementedError("must be implemented by subclass")
+        """Returns a dictionary mapping contents manager event actions to a handler (callable).
+
+        Returns a dictionary whose keys are contents manager event actions and whose values are callables
+        invoked upon receipt of an event of the same action. The callable accepts the body of the event as
+        its only argument. To ignore an event action, set the value to `None`.
+        """
+        pass
 
 
 class ArbitraryFileIdManager(BaseFileIdManager):
@@ -120,14 +247,21 @@ class ArbitraryFileIdManager(BaseFileIdManager):
         self.con.execute("PRAGMA journal_mode = WAL")
         self.con.execute(
             "CREATE TABLE IF NOT EXISTS Files("
-            "id INTEGER PRIMARY KEY AUTOINCREMENT, "
+            "id TEXT PRIMARY KEY NOT NULL, "
             "path TEXT NOT NULL UNIQUE"
             ")"
         )
         self.con.execute("CREATE INDEX IF NOT EXISTS ix_Files_path ON Files (path)")
         self.con.commit()
 
-    def index(self, path: str) -> int:
+    def _create(self, path: str) -> str:
+        path = self._normalize_path(path)
+        id = self._uuid()
+        self.con.execute("INSERT INTO Files (id, path) VALUES (?, ?)", (id, path))
+        return id
+
+    def index(self, path: str) -> str:
+        path = self._normalize_path(path)
         row = self.con.execute("SELECT id FROM Files WHERE path = ?", (path,)).fetchone()
         existing_id = row and row[0]
 
@@ -135,38 +269,51 @@ class ArbitraryFileIdManager(BaseFileIdManager):
             return existing_id
 
         # create new record
-        cursor = self.con.execute("INSERT INTO Files (path) VALUES (?)", (path,))
+        id = self._create(path)
         self.con.commit()
-        return cursor.lastrowid  # type:ignore
+        return id
 
-    def get_id(self, path: str) -> Optional[int]:
+    def get_id(self, path: str) -> Optional[str]:
+        path = self._normalize_path(path)
         row = self.con.execute("SELECT id FROM Files WHERE path = ?", (path,)).fetchone()
         return row and row[0]
 
-    def get_path(self, id: Union[int, str]) -> Optional[int]:
+    def get_path(self, id: str) -> Optional[str]:
         row = self.con.execute("SELECT path FROM Files WHERE id = ?", (id,)).fetchone()
-        return row and row[0]
+        path = row and row[0]
+        return self._from_normalized_path(path)
 
     def move(self, old_path: str, new_path: str) -> None:
+        old_path = self._normalize_path(old_path)
+        new_path = self._normalize_path(new_path)
         row = self.con.execute("SELECT id FROM Files WHERE path = ?", (old_path,)).fetchone()
         id = row and row[0]
 
         if id:
             self.con.execute("UPDATE Files SET path = ? WHERE path = ?", (new_path, old_path))
+            self._move_recursive(old_path, new_path, "/")
         else:
-            cursor = self.con.execute("INSERT INTO Files (path) VALUES (?)", (new_path,))
-            id = cursor.lastrowid
+            id = self._create(new_path)
 
         self.con.commit()
         return id
 
-    def copy(self, from_path: str, to_path: str) -> Optional[int]:
-        cursor = self.con.execute("INSERT INTO Files (path) VALUES (?)", (to_path,))
+    def copy(self, from_path: str, to_path: str) -> Optional[str]:
+        from_path = self._normalize_path(from_path)
+        to_path = self._normalize_path(to_path)
+
+        id = self._create(to_path)
+        self._copy_recursive(from_path, to_path, "/")
+
         self.con.commit()
-        return cursor.lastrowid
+        return id
 
     def delete(self, path: str) -> None:
+        path = self._normalize_path(path)
+
         self.con.execute("DELETE FROM Files WHERE path = ?", (path,))
+        self._delete_recursive(path, "/")
+
         self.con.commit()
 
     def save(self, path: str) -> None:
@@ -205,16 +352,15 @@ class LocalFileIdManager(BaseFileIdManager):
     performed during a method's procedure body.
     """
 
-    autosync_interval = Int(
-        default_value=5,
-        help=(
-            "How often to automatically invoke `sync_all()` when calling `get_path()`, in seconds. "
-            "Set to 0 to force `sync_all()` to always be called when calling `get_path()`. "
-            "Set to a negative value to disable autosync for `get_path()`."
-            "Defaults to 5."
-        ),
-        config=True,
-    )
+    @validate("root_dir")
+    def _validate_root_dir(self, proposal):
+        if proposal["value"] is None:
+            raise TraitError(f"LocalFileIdManager : {proposal['trait'].name} must not be None")
+        if not os.path.isabs(proposal["value"]):
+            raise TraitError(
+                f"LocalFileIdManager : {proposal['trait'].name} must be an absolute path"
+            )
+        return proposal["value"]
 
     def __init__(self, *args, **kwargs):
         # pass args and kwargs to parent Configurable
@@ -232,7 +378,7 @@ class LocalFileIdManager(BaseFileIdManager):
         self.con.execute("PRAGMA journal_mode = WAL")
         self.con.execute(
             "CREATE TABLE IF NOT EXISTS Files("
-            "id INTEGER PRIMARY KEY AUTOINCREMENT, "
+            "id TEXT PRIMARY KEY NOT NULL, "
             # uniqueness constraint relaxed here because we need to keep records
             # of deleted files which may occupy same path
             "path TEXT NOT NULL, "
@@ -248,6 +394,13 @@ class LocalFileIdManager(BaseFileIdManager):
         self.con.execute("CREATE INDEX IF NOT EXISTS ix_Files_is_dir ON Files (is_dir)")
         self.con.commit()
 
+    def _normalize_path(self, path):
+        path = super()._normalize_path(path)
+        path = os.path.normcase(path)
+        path = os.path.normpath(path)
+
+        return path
+
     def _index_all(self):
         """Recursively indexes all directories under the server root."""
         self._index_dir_recursively(self.root_dir, self._stat(self.root_dir))
@@ -377,7 +530,7 @@ class LocalFileIdManager(BaseFileIdManager):
 
         Returns
         -------
-        id : int, optional
+        id : str, optional
             ID of the file if it is a real file (not a symlink) and it was
             previously indexed. None otherwise.
 
@@ -414,14 +567,6 @@ class LocalFileIdManager(BaseFileIdManager):
 
         return id
 
-    def _normalize_path(self, path):
-        """Normalizes a given file path."""
-        if not os.path.isabs(path):
-            path = os.path.join(self.root_dir, path)
-        path = os.path.normcase(path)
-        path = os.path.normpath(path)
-        return path
-
     def _parse_raw_stat(self, raw_stat):
         """Accepts an `os.stat_result` object and returns a `StatStruct`
         object."""
@@ -462,12 +607,12 @@ class LocalFileIdManager(BaseFileIdManager):
         dangerous and may throw a runtime error if the file is not guaranteed to
         have a unique `ino`.
         """
-        cursor = self.con.execute(
-            "INSERT INTO Files (path, ino, crtime, mtime, is_dir) VALUES (?, ?, ?, ?, ?)",
-            (path, stat_info.ino, stat_info.crtime, stat_info.mtime, stat_info.is_dir),
+        id = self._uuid()
+        self.con.execute(
+            "INSERT INTO Files (id, path, ino, crtime, mtime, is_dir) VALUES (?, ?, ?, ?, ?, ?)",
+            (id, path, stat_info.ino, stat_info.crtime, stat_info.mtime, stat_info.is_dir),
         )
-
-        return cursor.lastrowid
+        return id
 
     def _update(self, id, stat_info=None, path=None):
         """Updates a record given its file ID and stat info.
@@ -505,18 +650,6 @@ class LocalFileIdManager(BaseFileIdManager):
             )
             return
 
-    def sync_all(self):
-        """
-        Syncs Files table with the filesystem and ensures that the correct path
-        is associated with each file ID.
-
-        Notes
-        -----
-        - Identical to `_sync_all()`, but is public and commits the transaction
-        upon invocation. See `_sync_all()` for more implementation details."""
-        self._sync_all()
-        self.con.commit()
-
     def index(self, path, stat_info=None, commit=True):
         """Returns the file ID for the file at `path`, creating a new file ID if
         one does not exist. Returns None only if file does not exist at path."""
@@ -562,20 +695,28 @@ class LocalFileIdManager(BaseFileIdManager):
 
         Notes
         -----
-        - To force syncing when calling `get_path()`, call `sync_all()` manually
+        - To force syncing when calling `get_path()`, call `_sync_all()` manually
         prior to calling `get_path()`.
         """
-        if (
-            self.autosync_interval >= 0
-            and (time.time() - self._last_sync) >= self.autosync_interval
-        ):
-            self._sync_all()
-
+        # optimistic approach: first check to see if path was not yet moved
         row = self.con.execute("SELECT path, ino FROM Files WHERE id = ?", (id,)).fetchone()
 
-        # if no record associated with ID, return None
-        if row is None:
+        # if file ID does not exist, return None
+        if not row:
             return None
+
+        path, ino = row
+        stat_info = self._stat(path)
+
+        if stat_info and ino == stat_info.ino and self._check_timestamps(stat_info):
+            # if file already exists at path and the ino and timestamps match,
+            # then return the correct path immediately (best case)
+            return self._from_normalized_path(path)
+
+        # otherwise, try again after calling _sync_all() to sync the Files table to the file tree
+        self._sync_all()
+        row = self.con.execute("SELECT path, ino FROM Files WHERE id = ?", (id,)).fetchone()
+        # file ID already guaranteed to exist from previous check
         path, ino = row
 
         # if file no longer exists at path, return None
@@ -588,28 +729,8 @@ class LocalFileIdManager(BaseFileIdManager):
         if ino != stat_info.ino or not self._check_timestamps(stat_info):
             return None
 
-        # if path is not relative to `self.root_dir`, return None.
-        if os.path.commonpath([self.root_dir, path]) != self.root_dir:
-            return None
-
-        # finally, convert the path to a relative one.
-        path = os.path.relpath(path, self.root_dir)
-        return path
-
-    def _move_recursive(self, old_path, new_path):
-        """Updates path of all indexed files prefixed with `old_path` and
-        replaces the prefix with `new_path`."""
-        old_path_glob = os.path.join(old_path, "*")
-        records = self.con.execute(
-            "SELECT id, path FROM Files WHERE path GLOB ?", (old_path_glob,)
-        ).fetchall()
-
-        for record in records:
-            id, old_recpath = record
-            new_recpath = os.path.join(new_path, os.path.relpath(old_recpath, start=old_path))
-
-            # only update path, not stat info
-            self._update(id, path=new_recpath)
+        # finally, convert the path to a relative one and return it
+        return self._from_normalized_path(path)
 
     @log(
         lambda self, old_path, new_path: f"Updating index following move from {old_path} to {new_path}.",
@@ -644,6 +765,22 @@ class LocalFileIdManager(BaseFileIdManager):
             self.con.commit()
             return id
 
+    def _copy_recursive(self, from_path: str, to_path: str, _: str = "") -> None:
+        """Copy all children of a given directory at `from_path` to a new
+        directory at `to_path`. Inserts stat_info with record."""
+        from_path_glob = os.path.join(from_path, "*")
+        records = self.con.execute(
+            "SELECT path FROM Files WHERE path GLOB ?", (from_path_glob,)
+        ).fetchall()
+
+        for record in records:
+            (from_recpath,) = record
+            to_recpath = os.path.join(to_path, os.path.relpath(from_recpath, start=from_path))
+            stat_info = self._stat(to_recpath)
+            if not stat_info:
+                continue
+            self._create(to_recpath, stat_info)
+
     @log(
         lambda self, from_path, to_path: f"Indexing {to_path} following copy from {from_path}.",
         lambda self, from_path, to_path: f"Successfully indexed {to_path} following copy from {from_path}.",
@@ -658,28 +795,7 @@ class LocalFileIdManager(BaseFileIdManager):
         to_path = self._normalize_path(to_path)
 
         if os.path.isdir(to_path):
-            from_path_glob = os.path.join(from_path, "*")
-            records = self.con.execute(
-                "SELECT path FROM Files WHERE path GLOB ?", (from_path_glob,)
-            ).fetchall()
-            for record in records:
-                if not record:
-                    continue
-                (from_recpath,) = record
-                to_recpath = os.path.join(to_path, os.path.relpath(from_recpath, start=from_path))
-                stat_info = self._stat(to_recpath)
-                if not stat_info:
-                    continue
-                self.con.execute(
-                    "INSERT INTO FILES (path, ino, crtime, mtime, is_dir) VALUES (?, ?, ?, ?, ?)",
-                    (
-                        to_recpath,
-                        stat_info.ino,
-                        stat_info.crtime,
-                        stat_info.mtime,
-                        stat_info.is_dir,
-                    ),
-                )
+            self._copy_recursive(from_path, to_path)
 
         self.index(from_path, commit=False)
         # transaction committed in index()
@@ -695,8 +811,7 @@ class LocalFileIdManager(BaseFileIdManager):
         path = self._normalize_path(path)
 
         if os.path.isdir(path):
-            path_glob = os.path.join(path, "*")
-            self.con.execute("DELETE FROM Files WHERE path GLOB ?", (path_glob,))
+            self._delete_recursive(path)
 
         self.con.execute("DELETE FROM Files WHERE path = ?", (path,))
         self.con.commit()
@@ -719,7 +834,6 @@ class LocalFileIdManager(BaseFileIdManager):
         row = self.con.execute(
             "SELECT id FROM Files WHERE ino = ? AND path = ?", (stat_info.ino, path)
         ).fetchone()
-
         # if no record exists, return early
         if row is None:
             return
