@@ -1,3 +1,4 @@
+import ntpath
 import os
 from unittest.mock import patch
 
@@ -78,21 +79,51 @@ def get_path_nosync(fid_manager, id):
 
 
 def test_validates_root_dir(fid_db_path):
+    root_dir = "s3://bucket"
     with pytest.raises(TraitError, match="must be an absolute path"):
-        LocalFileIdManager(root_dir=os.path.join("some", "rel", "path"), db_path=fid_db_path)
-    with pytest.raises(TraitError, match="must be an absolute path"):
-        ArbitraryFileIdManager(root_dir=os.path.join("some", "rel", "path"), db_path=fid_db_path)
+        LocalFileIdManager(root_dir=root_dir, db_path=fid_db_path)
+    # root_dir can be relative for ArbitraryFileIdManager instances (and None)
+    afm = ArbitraryFileIdManager(root_dir=root_dir, db_path=fid_db_path)
+    assert afm.root_dir == root_dir
+    afm2 = ArbitraryFileIdManager(root_dir=None, db_path=fid_db_path)
+    assert afm2.root_dir is None
 
 
-def test_validates_db_path(jp_root_dir):
+def test_validates_db_path(jp_root_dir, any_fid_manager_class):
     with pytest.raises(TraitError, match="must be an absolute path"):
-        LocalFileIdManager(root_dir=str(jp_root_dir), db_path=os.path.join("some", "rel", "path"))
-    with pytest.raises(TraitError, match="must be an absolute path"):
-        ArbitraryFileIdManager(
+        any_fid_manager_class(
             root_dir=str(jp_root_dir), db_path=os.path.join("some", "rel", "path")
         )
 
 
+def test_different_roots(
+    any_fid_manager_class, fid_db_path, jp_root_dir, test_path, test_path_child
+):
+    """Assert that default FIM implementations assign the same file the same
+    file ID agnostic of contents manager root."""
+    fid_manager_1 = any_fid_manager_class(db_path=fid_db_path, root_dir=str(jp_root_dir))
+    fid_manager_2 = any_fid_manager_class(
+        db_path=fid_db_path, root_dir=str(jp_root_dir / test_path)
+    )
+
+    id_1 = fid_manager_1.index(test_path_child)
+    id_2 = fid_manager_2.index(os.path.basename(test_path_child))
+
+    assert id_1 == id_2
+
+
+def test_different_roots_arbitrary(fid_db_path):
+    """Assert that ArbitraryFileIdManager assigns the same file the same file ID
+    agnostic of contents manager root, even if non-local."""
+    manager_1 = ArbitraryFileIdManager(db_path=fid_db_path, root_dir="s3://bucket")
+    manager_2 = ArbitraryFileIdManager(db_path=fid_db_path, root_dir="s3://bucket/folder")
+
+    id_1 = manager_1.index("folder/child")
+    id_2 = manager_2.index("child")
+
+    assert id_1 == id_2
+
+
 def test_index(any_fid_manager, test_path):
     id = any_fid_manager.index(test_path)
     assert id is not None
@@ -246,6 +277,46 @@ def test_get_id_oob_move_new_file_at_old_path(fid_manager, old_path, new_path, f
     assert fid_manager.get_id(other_path) == other_id
 
 
+def test_get_path_arbitrary_preserves_path(arbitrary_fid_manager):
+    """Tests whether ArbitraryFileIdManager always preserves the file paths it
+    receives."""
+    path = "AbCd.txt"
+    id = arbitrary_fid_manager.index(path)
+    assert path == arbitrary_fid_manager.get_path(id)
+
+
+@patch("os.path.sep", new="\\")
+@patch("os.path.relpath", new=ntpath.relpath)
+@patch("os.path.normpath", new=ntpath.normpath)
+@patch("os.path.join", new=ntpath.join)
+def test_get_path_returns_api_path(jp_root_dir, fid_db_path):
+    """Tests whether get_path() method always returns an API path, i.e. one
+    relative to the server root and one delimited by forward slashes (even if
+    os.path.sep = "\\")."""
+    test_path = "a\\b\\c"
+    expected_path = "a/b/c"
+    manager = ArbitraryFileIdManager(
+        root_dir=ntpath.join("c:", ntpath.normpath(str(jp_root_dir))), db_path=fid_db_path
+    )
+
+    id = manager.index(test_path)
+    path = manager.get_path(id)
+    assert path == expected_path
+
+
+def test_optimistic_get_path(fid_manager, test_path, test_path_child):
+    """_sync_all() should never be called in the best case, in which no paths
+    were moved out-of-band."""
+    with patch.object(fid_manager, "_sync_all") as mock:
+        id_1 = fid_manager.index(test_path)
+        id_2 = fid_manager.index(test_path_child)
+
+        fid_manager.get_path(id_1)
+        fid_manager.get_path(id_2)
+
+        mock.assert_not_called()
+
+
 def test_get_path_oob_move(fid_manager, old_path, new_path, fs_helpers):
     id = fid_manager.index(old_path)
     fs_helpers.move(old_path, new_path)
@@ -282,7 +353,6 @@ def test_get_path_oob_move_back_to_original_path(fid_manager, old_path, new_path
 
     assert fid_manager.get_path(id) == new_path
     fs_helpers.move(new_path, old_path)
-    fid_manager.sync_all()
     assert fid_manager.get_path(id) == old_path
 
 
@@ -327,7 +397,7 @@ def test_move_unindexed(any_fid_manager, old_path, new_path, fs_helpers):
 
     assert id is not None
     assert any_fid_manager.get_id(old_path) is None
-    assert any_fid_manager.get_id(new_path) is id
+    assert any_fid_manager.get_id(new_path) == id
     assert any_fid_manager.get_path(id) == new_path
 
 
@@ -356,7 +426,7 @@ def test_disjoint_move_indexed(any_fid_manager, old_path, new_path, fs_helpers):
 
 
 def test_move_recursive(
-    fid_manager,
+    any_fid_manager,
     old_path,
     old_path_child,
     old_path_grandchild,
@@ -365,18 +435,18 @@ def test_move_recursive(
     new_path_grandchild,
     fs_helpers,
 ):
-    parent_id = fid_manager.index(old_path)
-    child_id = fid_manager.index(old_path_child)
-    grandchild_id = fid_manager.index(old_path_grandchild)
+    parent_id = any_fid_manager.index(old_path)
+    child_id = any_fid_manager.index(old_path_child)
+    grandchild_id = any_fid_manager.index(old_path_grandchild)
 
     fs_helpers.move(old_path, new_path)
-    fid_manager.move(old_path, new_path)
+    any_fid_manager.move(old_path, new_path)
 
     # we avoid using get_id() here as it auto-corrects wrong path updates via
     # its out-of-band move detection logic. too smart for its own good!
-    assert get_id_nosync(fid_manager, new_path) == parent_id
-    assert get_id_nosync(fid_manager, new_path_child) == child_id
-    assert get_id_nosync(fid_manager, new_path_grandchild) == grandchild_id
+    assert get_id_nosync(any_fid_manager, new_path) == parent_id
+    assert get_id_nosync(any_fid_manager, new_path_child) == child_id
+    assert get_id_nosync(any_fid_manager, new_path_grandchild) == grandchild_id
 
 
 def test_copy(any_fid_manager, old_path, new_path, fs_helpers):
@@ -390,7 +460,7 @@ def test_copy(any_fid_manager, old_path, new_path, fs_helpers):
 
 
 def test_copy_recursive(
-    fid_manager,
+    any_fid_manager,
     old_path,
     old_path_child,
     old_path_grandchild,
@@ -399,16 +469,16 @@ def test_copy_recursive(
     new_path_grandchild,
     fs_helpers,
 ):
-    fid_manager.index(old_path)
-    fid_manager.index(old_path_child)
-    fid_manager.index(old_path_grandchild)
+    any_fid_manager.index(old_path)
+    any_fid_manager.index(old_path_child)
+    any_fid_manager.index(old_path_grandchild)
 
     fs_helpers.copy(old_path, new_path)
-    fid_manager.copy(old_path, new_path)
+    any_fid_manager.copy(old_path, new_path)
 
-    assert fid_manager.get_id(new_path) is not None
-    assert fid_manager.get_id(new_path_child) is not None
-    assert fid_manager.get_id(new_path_grandchild) is not None
+    assert any_fid_manager.get_id(new_path) is not None
+    assert any_fid_manager.get_id(new_path_child) is not None
+    assert any_fid_manager.get_id(new_path_grandchild) is not None
 
 
 def test_delete(any_fid_manager, test_path, fs_helpers):
@@ -421,14 +491,14 @@ def test_delete(any_fid_manager, test_path, fs_helpers):
     assert any_fid_manager.get_path(id) is None
 
 
-def test_delete_recursive(fid_manager, test_path, test_path_child, fs_helpers):
-    fid_manager.index(test_path)
-    fid_manager.index(test_path_child)
+def test_delete_recursive(any_fid_manager, test_path, test_path_child, fs_helpers):
+    any_fid_manager.index(test_path)
+    any_fid_manager.index(test_path_child)
 
     fs_helpers.delete(test_path)
-    fid_manager.delete(test_path)
+    any_fid_manager.delete(test_path)
 
-    assert fid_manager.get_id(test_path_child) is None
+    assert any_fid_manager.get_id(test_path_child) is None
 
 
 def test_save(any_fid_manager, test_path, fs_helpers):
@@ -438,39 +508,3 @@ def test_save(any_fid_manager, test_path, fs_helpers):
     any_fid_manager.save(test_path)
 
     assert any_fid_manager.get_id(test_path) == id
-
-
-def test_autosync_gt_0(fid_manager, old_path, new_path, fs_helpers):
-    fid_manager.autosync_interval = 10
-    id = fid_manager.index(old_path)
-    fid_manager.sync_all()
-    fs_helpers.move(old_path, new_path)
-
-    assert fid_manager.get_path(id) != new_path
-    with patch("time.time") as mock_time:
-        mock_time.return_value = fid_manager._last_sync + 999
-        assert fid_manager.get_path(id) == new_path
-
-
-def test_autosync_eq_0(fid_manager, old_path, new_path, fs_helpers):
-    fid_manager.autosync_interval = 0
-    id = fid_manager.index(old_path)
-    fid_manager.sync_all()
-    fs_helpers.move(old_path, new_path)
-
-    assert fid_manager.get_path(id) == new_path
-
-
-def test_autosync_lt_0(fid_manager, old_path, new_path, fs_helpers):
-    fid_manager.autosync_interval = -10
-    id = fid_manager.index(old_path)
-    fid_manager.sync_all()
-    fs_helpers.move(old_path, new_path)
-
-    assert fid_manager.get_path(id) != new_path
-    with patch("time.time") as mock_time:
-        mock_time.return_value = fid_manager._last_sync + 999
-        assert fid_manager.get_path(id) != new_path
-
-    fid_manager.sync_all()
-    assert fid_manager.get_path(id) == new_path
